package renderer;

import canvases.CalculatingFPS;
import canvases.MainMenu;
import canvases.SceneCanvas;
import model.Controller;
import model.Scene;
import lwjglutils.*;


import org.lwjgl.BufferUtils;
import org.lwjgl.glfw.GLFWMouseButtonCallback;
import org.lwjgl.glfw.GLFWWindowSizeCallback;

import transforms.*;


import java.io.IOException;
import java.nio.DoubleBuffer;
import java.util.ArrayList;
import java.util.List;

import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.*;


public class ShaderRenderer extends AbstractRenderer {
    private List<IShader> shaders;
    private IShader activeShader;


    private final float speed = .1f;
    private final List<OGLTexture2D> texture = new ArrayList<>();
    private final Scene scene;
    private OGLBuffers buffers;
    private float h = 0;
    private int sceneId = 0;
    long lastTime = 0;
    MainMenuShader mainMenu;


    public ShaderRenderer(Scene scene, Controller controller) {
        super(controller);
        this.scene = scene;
        glfwWindowSizeCallback = new GLFWWindowSizeCallback() {
            @Override
            public void invoke(long window, int w, int h) {
                if (w > 0 && h > 0) {
                    width = w;
                    height = h;
                    canvas.resize(width, height);
                }
            }
        };

        /*used default glfwKeyCallback */

        glfwMouseButtonCallback = new GLFWMouseButtonCallback() {
            @Override
            public void invoke(long window, int button, int action, int mods) {
                double[] xPos = new double[1];
                double[] yPos = new double[1];
                glfwGetCursorPos(window, xPos, yPos);
                float x = (float) xPos[0] / width * 2 - 1;
                float y = (float) yPos[0] / height * 2 - 1;
                y *= -1;
                canvas.checkButtons(button, action, x, y);
                DoubleBuffer xBuffer = BufferUtils.createDoubleBuffer(1);
                DoubleBuffer yBuffer = BufferUtils.createDoubleBuffer(1);
                glfwGetCursorPos(window, xBuffer, yBuffer);
                double xD = xBuffer.get(0);
                double yD = yBuffer.get(0);

                if (action == GLFW_PRESS) {
                    controller.mouseBtnPressed(button, xD, yD);
                }

                if (action == GLFW_RELEASE) {
                    controller.mouseBtnReleased(button, xD, yD);
                }
            }
        };
    }

    @Override
    public void init() {
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        createBuffers();
        shaders = new ArrayList<>();
        int shader1 = ShaderUtils.loadProgram("/RayMarching.vert",
                "/RayMarching.frag",
                null, null, null, null);
        int shader2 = ShaderUtils.loadProgram("/RayMarching.vert",
                "/RayMarching2.frag",
                null, null, null, null);
        int shader3 = ShaderUtils.loadProgram("/RayMarching.vert",
                "/RayMarching3.frag",
                null, null, null, null);
        int mainMenuShader = ShaderUtils.loadProgram("/RayMarching.vert",
                "/MainMenu.frag",
                null, null, null, null);
        Shader s1 = new Shader(shader1);
        Shader s2 = new Shader(shader2);
        Shader s3 = new Shader(shader3);
        mainMenu = new MainMenuShader(mainMenuShader);
        shaders.add(mainMenu);
        shaders.add(s1);
        shaders.add(s2);
        shaders.add(s3);
        activeShader = mainMenu;
        glUseProgram(mainMenuShader);
        try {
            texture.add(new OGLTexture2D("ground.jpg"));
            texture.add(new OGLTexture2D("screen.jpg"));
            texture.add(new OGLTexture2D("wood.png"));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        canvas = new MainMenu(width, height, this);
        canvases.add(canvas);
        canvases.add(new SceneCanvas(width, height, this));
    }

    @Override
    public void display() {
        if (sceneId == 0) {
            long time = System.currentTimeMillis();
            float deltaTime = (time - lastTime) / 1000f;
            if (deltaTime > 1) {
                deltaTime = 0;
            }
            lastTime = time;
            mainMenu.addTime(deltaTime);
            scene.rotateAround(deltaTime);
        }
        if(sceneId == 2){
            scene.changeColor();
        }
        ((CalculatingFPS) canvas).calculateFPS();
        glViewport(0, 0, width, height);

        scene.move(controller.isMoving());
        activeShader.render();
        canvas.render();
    }


    void createBuffers() {
        float[] vertexBufferData = {
                -1, -1, 0.7f, 0, 0,
                1, -1, 0, 0.7f, 0,
                -1, 1, 0, 0, 0.7f,
                1, 1, 0.7f, 0, 0
        };
        int[] indexBufferData = {0, 1, 2, 2, 1, 3};

        OGLBuffers.Attrib[] attributes = {
                new OGLBuffers.Attrib("inPosition", 2), // 2 floats
                new OGLBuffers.Attrib("inColor", 3) // 3 floats
        };
        buffers = new OGLBuffers(vertexBufferData, attributes,
                indexBufferData);
    }

    public void setSceneId(int sceneId) {

        this.sceneId = sceneId;
        activeShader = shaders.get(sceneId);
        if (sceneId == 0)
            canvas = canvases.get(sceneId);
        else
            canvas = canvases.get(1);
        scene.setMultipleLights(sceneId == 2);
        glUseProgram(shaders.get(sceneId).getShader());
    }


    private interface IShader {
        void render();

        int getShader();
    }

    private class MainMenuShader implements IShader {
        @Override
        public int getShader() {
            return shader;
        }

        public void addTime(float time) {
            this.time = increasing ? this.time + time : this.time - time;
            if (this.time > 5f && increasing) {
                increasing = false;
            } else {
                if (this.time < 1f && !increasing) {
                    increasing = true;
                }
            }
        }

        private boolean increasing = true;

        private final int shader;
        private final int widthLoc;
        private final int heightLoc;
        private final int camPosLoc;
        private final int lookAtLoc;
        private final int timeLoc;

        private float time = 0f;

        public MainMenuShader(int shader) {
            this.shader = shader;
            widthLoc = glGetUniformLocation(shader, "width");
            heightLoc = glGetUniformLocation(shader, "height");
            camPosLoc = glGetUniformLocation(shader, "camPos");
            lookAtLoc = glGetUniformLocation(shader, "lookAt");
            timeLoc = glGetUniformLocation(shader, "time");
        }

        public void render() {
            Camera camera = scene.getCamera();
            float x = (float) camera.getPosition().getX();
            float y = (float) camera.getPosition().getY();
            float z = (float) camera.getPosition().getZ();
            glUseProgram(shader);
            glUniform1f(widthLoc, width);
            glUniform1f(heightLoc, height);
            glUniform3f(camPosLoc, x, y, z);
            glUniformMatrix4fv(lookAtLoc, false, scene.getCamera().getViewMatrix().floatArray());
            glUniform1f(timeLoc, time);
            h += 0.01f;
            if (h > 1) h = 0;
            buffers.draw(GL_TRIANGLES, shader);
        }
    }

    private class Shader implements IShader {
        @Override
        public int getShader() {
            return shader;
        }

        private final int shader;
        private final int widthLoc;
        private final int heightLoc;
        private final int camPosLoc;
        private final int lookAtLoc;
        private final int lightPosLoc;
        private final int lightColLoc;
        private final int numOfLightsLoc;


        public Shader(int shader) {
            this.shader = shader;
            widthLoc = glGetUniformLocation(shader, "width");
            heightLoc = glGetUniformLocation(shader, "height");
            camPosLoc = glGetUniformLocation(shader, "camPos");
            lookAtLoc = glGetUniformLocation(shader, "lookAt");
            lightPosLoc = glGetUniformLocation(shader, "lightPos");
            lightColLoc = glGetUniformLocation(shader, "lightCol");
            numOfLightsLoc = glGetUniformLocation(shader, "numOfLights");
        }

        public void render() {
            Camera camera = scene.getCamera();
            float x = (float) camera.getPosition().getX();
            float y = (float) camera.getPosition().getY();
            float z = (float) camera.getPosition().getZ();

            glUseProgram(shader);
            glUniform1f(widthLoc, width);
            glUniform1f(heightLoc, height);
            glUniform1i(numOfLightsLoc, scene.getNumOfLights());
            glUniform3f(camPosLoc, x, y, z);
            glUniformMatrix4fv(lookAtLoc, false, scene.getCamera().getViewMatrix().floatArray());
            glUniform3fv(lightPosLoc, scene.getLightPos());
            glUniform3fv(lightColLoc, scene.getLightCol());

            for (int i = 0; i < texture.size(); i++) {
                texture.get(i).bind(shader, "texture" + (i + 1), i);
            }

            buffers.draw(GL_TRIANGLES, shader);
        }
    }
}
