package ui.components;

import lwjglutils.OGLBuffers;
import lwjglutils.ShaderUtils;

import model.Canvas;
import transforms.Vec2D;
import ui.utils.UIButtonListener;
import ui.utils.UITextAlign;
import ui.utils.UITextProvider;

import java.awt.*;

import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.*;


public class UIButton extends UIComponent implements UITextProvider {
    private final int shader;
    private float r,g,b;
    private boolean autoResize = true, pressed = false;

    private UIButtonListener listener;
    private OGLBuffers buffers;
    private final int dimensionLoc, radiusLoc, uvLoc, positionLoc, borderWidthLoc, colorLoc;
    private float centeredPosX, centeredPosY, correctedX;
    private float borderWidth = 0.01f;
    private float radius = 0.2f;


    private UILabel label;

    /**
     * @param componentWidth šířka
     * @param componentHeight výška
     * @param posX pozice x v ndc
     * @param posY pozice y v ndc
     * @param color barva
     * @param canvas kanvas
     * @param text text
     */
    public UIButton(float componentWidth, float componentHeight, float posX, float posY, Color color, Canvas canvas, String text) {
        super(componentWidth, componentHeight, posX, posY, canvas);
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;

        shader = ShaderUtils.loadProgram("/Button.vert",
                "/Button.frag",
                null,null,null,null);
        glUseProgram(this.shader);
        dimensionLoc = glGetUniformLocation(shader, "dimensions");
        radiusLoc = glGetUniformLocation(shader, "radius");
        uvLoc = glGetUniformLocation(shader, "uv");
        positionLoc = glGetUniformLocation(shader, "position");
        borderWidthLoc = glGetUniformLocation(shader, "borderWidth");
        colorLoc = glGetUniformLocation(shader, "color");

        centeredPosX = posX + componentWidth/2;
        centeredPosY = posY + correctedHeight/2;

        label = new UILabel(componentWidth, componentHeight, centeredPosX, centeredPosY, text, canvas, UITextAlign.CENTER);
        float aspect = (float)canvas.getWidth()/(float)canvas.getHeight();
        correctedX = centeredPosX * aspect;
    }
    public UIButton(float componentWidth, float componentHeight, float posX, float posY, Color color, Canvas canvas) {
        this(componentWidth, componentHeight, posX, posY, color, canvas, "");
    }
    @Override
    public void render() {
        glUseProgram(shader);

        glUniform2f(dimensionLoc, correctedWidth/2, correctedHeight/2);
        glUniform2f(uvLoc, canvas.getWidth(), canvas.getHeight());
        glUniform1f(radiusLoc, radius);
        glUniform2f(positionLoc, correctedX, centeredPosY);
        glUniform1f(borderWidthLoc, 0.01f);
        if (pressed) {
            glUniform3f(colorLoc, r/2, g/2, b/2);
        } else {
            glUniform3f(colorLoc, r, g, b);
        }

        buffers.draw(GL_TRIANGLES, shader);
        glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, 0);

    }

    @Override
    public void resize(int width, int height) {
        float aspectRatio = (float)width/(float)height;
        correctedHeight = this.componentHeight * aspectRatio;
        correctedWidth = this.componentWidth * aspectRatio;

        createBuffers();
        centeredPosY = y + correctedHeight/2;
        correctedX = centeredPosX * aspectRatio;
        label.setY(centeredPosY);
        label.resize(width, height);
    }

    public void setListener(UIButtonListener listener) {
        this.listener = listener;
    }
    public UIButtonListener getListener() {
        return listener;
    }
    public void removeListener() {
        listener = null;
    }
    public void checkButton(int button, int action, float posX, float posY) {

        if (listener != null) {
            if (sdRoundedBox(new Vec2D(posX-centeredPosX, posY-centeredPosY), new Vec2D(componentWidth/2, correctedHeight/2), 0.2f) < 0) {
                listener.performAction(button, action);
                pressed = true;
            }
        }
    }
    void createBuffers() {
        float[] vertexBufferData = {
                x, y, 	r, g, b,
                x+ componentWidth,  y,		r, g, b,
                x,  y+correctedHeight,		r, g, b,
                x+ componentWidth,  y+correctedHeight,		r, g, b
        };
        int[] indexBufferData = { 0, 1, 2, 2, 1, 3};

        OGLBuffers.Attrib[] attributes = {
                new OGLBuffers.Attrib("inPosition", 2), // 2 floats
                new OGLBuffers.Attrib("inColor", 3) // 3 floats
        };
        buffers = new OGLBuffers(vertexBufferData, attributes,
                indexBufferData);

    }
    public void setText(String text) {
        label.setText(text);
    }
    public void setColor(Color color) {
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;
    }
    public void setTextColor(Color color) {
        label.setTextColor(color);
    }

    /**
     * @param borderWidth values > 1
     */
    public void setBorderWidth(float borderWidth) {
        if (borderWidth < 1) {
            borderWidth = 1;
        }
        this.borderWidth = 1/borderWidth;
    }

    public void setRadius(float radius) {
        this.radius = radius;
    }

    private float sdRoundedBox(Vec2D point, Vec2D size, float radius)
    {
        Vec2D q = new Vec2D(Math.abs(point.getX())-size.getX()+radius, Math.abs(point.getY())-size.getY()+radius);
        return (float)(Math.min(Math.max(q.getX(),q.getY()),0.0) + new Vec2D(Math.max(q.getX(),0.0), Math.max(q.getY(),0.0)).length() - radius);
    }

    public void release() {
        pressed = false;
    }

    @Override
    public void renderText() {
        label.renderText();
    }
}
