package ui.components;

import lwjglutils.OGLBuffers;
import lwjglutils.ShaderUtils;
import model.Canvas;
import org.lwjgl.glfw.GLFW;
import transforms.Vec2D;
import ui.utils.UITextAlign;
import ui.utils.UITextProvider;

import java.awt.*;


import static org.lwjgl.opengl.GL11.GL_TRIANGLES;
import static org.lwjgl.opengl.GL11.GL_UNSIGNED_SHORT;
import static org.lwjgl.opengl.GL11.glDrawElements;

import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL20.glUniform3f;

public class UIInputField extends UIComponent implements UITextProvider {
    private final int shader;
    private float r,g,b;
    private boolean autoResize = true;

    private OGLBuffers buffers;
    private final int dimensionLoc, radiusLoc, uvLoc, positionLoc, borderWidthLoc, colorLoc;
    private float centeredPosX, centeredPosY, correctedX;
    private float borderWidth = 0.05f;
    private float radius = 0.2f;
    private String text = "";
    private String placeholder = "";


    private UILabel label;

    /**
     * @param componentWidth width of the button
     * @param componentHeight height of the button
     * @param posX position x in ndc
     * @param posY position y in ndc
     * @param color color of the button
     * @param canvas canvas
     * @param placeholder text to be displayed
     */
    public UIInputField(float componentWidth, float componentHeight, float posX, float posY, Color color, model.Canvas canvas, String placeholder) {
        super(componentWidth, componentHeight, posX, posY, canvas);
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;

        shader = ShaderUtils.loadProgram("/Button.vert",
                "/InputField.frag",
                null,null,null,null);
        glUseProgram(this.shader);
        dimensionLoc = glGetUniformLocation(shader, "dimensions");
        radiusLoc = glGetUniformLocation(shader, "radius");
        uvLoc = glGetUniformLocation(shader, "uv");
        positionLoc = glGetUniformLocation(shader, "position");
        borderWidthLoc = glGetUniformLocation(shader, "borderWidth");
        colorLoc = glGetUniformLocation(shader, "color");

        centeredPosX = posX + componentWidth/2;
        centeredPosY = posY + correctedHeight/2;

        label = new UILabel(componentWidth, componentHeight, x+borderWidth, centeredPosY, placeholder, canvas, UITextAlign.CENTER_Y);
        this.placeholder = placeholder;
        float aspect = (float)canvas.getWidth()/(float)canvas.getHeight();
        correctedX = centeredPosX * aspect;
        System.out.println("centeredPosY: " + centeredPosY+" " + "centeredPosX: " + centeredPosX);
    }
    public UIInputField(float componentWidth, float componentHeight, float posX, float posY, Color color, Canvas canvas) {
        this(componentWidth, componentHeight, posX, posY, color, canvas, "");
    }
    @Override
    public void render() {
        glUseProgram(shader);

        glUniform2f(dimensionLoc, correctedWidth/2, correctedHeight/2);
        glUniform2f(uvLoc, canvas.getWidth(), canvas.getHeight());
        glUniform1f(radiusLoc, radius);
        glUniform2f(positionLoc, correctedX, centeredPosY);
        glUniform1f(borderWidthLoc, borderWidth);
        glUniform3f(colorLoc, r, g, b);


        buffers.draw(GL_TRIANGLES, shader);
        glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, 0);

        glUseProgram(0);

    }


    @Override
    public void resize(int width, int height) {
        float aspectRatio = (float)width/(float)height;
        correctedHeight = this.componentHeight * aspectRatio;
        correctedWidth = this.componentWidth * aspectRatio;

        createBuffers();
        centeredPosY = y + correctedHeight/2;
        correctedX = centeredPosX * aspectRatio;
        label.setY(centeredPosY);
        label.resize(width, height);

        changeText();
    }


    void createBuffers() {
        float[] vertexBufferData = {
                x, y, 	r, g, b,
                x+ componentWidth,  y,		r, g, b,
                x,  y+correctedHeight,		r, g, b,
                x+ componentWidth,  y+correctedHeight,		r, g, b
        };
        int[] indexBufferData = { 0, 1, 2, 2, 1, 3};

        OGLBuffers.Attrib[] attributes = {
                new OGLBuffers.Attrib("inPosition", 2), // 2 floats
                new OGLBuffers.Attrib("inColor", 3) // 3 floats
        };
        buffers = new OGLBuffers(vertexBufferData, attributes,
                indexBufferData);

    }
    public void setText(String text) {
        this.text = text;
        changeText();
    }
    public void setTextInLabel(String text) {
        label.setText(text);
    }
    public String getText() {
        return text;
    }
    public void setColor(Color color) {
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;
    }
    public void setTextColor(Color color) {
        label.setTextColor(color);
    }

    /**
     * @param borderWidth values > 1
     */
    public void setBorderWidth(float borderWidth) {
        if (borderWidth < 1) {
            borderWidth = 1;
        }
        this.borderWidth = 1/borderWidth;
    }

    public void setRadius(float radius) {
        this.radius = radius;
    }

    private float sdRoundedBox(Vec2D point, Vec2D size, float radius)
    {
        Vec2D q = new Vec2D(Math.abs(point.getX())-size.getX()+radius, Math.abs(point.getY())-size.getY()+radius);
        return (float)(Math.min(Math.max(q.getX(),q.getY()),0.0) + new Vec2D(Math.max(q.getX(),0.0), Math.max(q.getY(),0.0)).length() - radius);
    }


    public boolean checkField(int button, int action, float posX, float posY) {
        if (sdRoundedBox(new Vec2D(posX-centeredPosX, posY-centeredPosY), new Vec2D(componentWidth/2, correctedHeight/2), 0.2f) < 0) {
            return true;
        }
        return false;
    }
    public void keyTyped(int codePoint) {
        text += (char)codePoint;
        changeText();
    }

    public void keyPressed(int key) {
        if (key == GLFW.GLFW_KEY_BACKSPACE) {
            if (text.length() > 0) {
                text = text.substring(0, text.length() - 1);
                changeText();
            }
        }
    }
    private void changeText() {
        if (text.length() == 0) {
            label.setText(placeholder);
            return;
        }
        float max = componentWidth - 2 * borderWidth;
        String truncatedText = text;

        while (label.getTextWidth(truncatedText) > max && truncatedText.length() > 0) {
            truncatedText = truncatedText.substring(1);
        }

        label.setText(truncatedText);
    }

    public void setPlaceholder(String placeholder) {
        this.placeholder = placeholder;
    }

    @Override
    public void renderText() {
        label.renderText();
    }
}
